/* usbhotkey: Turn USB keyboard events into X11 keyboard events.                   */
/* Copyright (C) 2007  Lars Krueger                                                */
/*                                                                                 */
/* This program is free software; you can redistribute it and/or modify            */
/* it under the terms of the GNU General Public License as published by            */
/* the Free Software Foundation; either version 2 of the License, or               */
/* (at your option) any later version.                                             */
/*                                                                                 */
/* This program is distributed in the hope that it will be useful,                 */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of                  */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   */
/* GNU General Public License for more details.                                    */
/*                                                                                 */
/* You should have received a copy of the GNU General Public License               */
/* along with this program; if not, write to the Free Software                     */
/* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA  */

/* The ruby specific code is taken from various tutorials and forum entries
 * from all over the net.
 *
 * All constructs are documented for the use with doxygen. It is recommended
 * to generate the html documentation if you want to extend usbhotkey
 * yourself.
 *
 * Feel free to send any improvements, patches or change requests to
 * lars_e_krueger@sourceforge.net
 *
 */

#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <config.h>
#include <stdarg.h>
#include <assert.h>

/* Undef the symbols that may collide with ruby. */
#undef PACKAGE_NAME
#undef PACKAGE_VERSION
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME

#include <hid.h>

#include <ruby.h>

#if RUBY_VERSION >= 19
/** Portable backtrace information. */
#define RB_ERRINFO rb_errinfo()
#else
#define RB_ERRINFO ruby_errinfo
#endif

#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>
#include <X11/XKBlib.h>

#include "constants.h"

/* Uncomment this to enable packet debugging. */
/* #define DEBUG_PACKET */

/** USB HID scancode not found on AT type keyboards.  */
#define unk 0


/** Number of entries to in usb2at_keymap */
#define USB2AT_ENTRIES 256

/** This table maps USB HID scancodes (the type that will be returned by USB
 * keyboards this utility interfaces with) to AT type scancodes.  Adding 8 to
 * these values will make them usable for looking up keysyms with Xkb.  This
 * is pulled verbatim (just renamed) from the linux kernel source: hid-input.c
 * */
static const KeyCode usb2at_keymap[USB2AT_ENTRIES] = {
  0, 0, 0, 0, 30, 48, 46, 32, 18, 33, 34, 35, 23, 36, 37, 38, 50, 49, 24, 25,
  16, 19, 31, 20, 22, 47, 17, 45, 21, 44, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 28,
  1, 14, 15, 57, 12, 13, 26, 27, 43, 43, 39, 40, 41, 51, 52, 53, 58, 59, 60,
  61, 62, 63, 64, 65, 66, 67, 68, 87, 88, 99, 70, 119, 110, 102, 104, 111,
  107, 109, 106, 105, 108, 103, 69, 98, 55, 74, 78, 96, 79, 80, 81, 75, 76,
  77, 71, 72, 73, 82, 83, 86, 127, 116, 117, 183, 184, 185, 186, 187, 188,
  189, 190, 191, 192, 193, 194, 134, 138, 130, 132, 128, 129, 131, 137, 133,
  135, 136, 113, 115, 114, unk, unk, unk, 121, unk, 89, 93, 124, 92, 94, 95,
  unk, unk, unk, 122, 123, 90, 91, 85, unk, unk, unk, unk, unk, unk, unk, 111,
  unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk,
  unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, 179, 180, unk, unk, unk,
  unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk,
  unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, unk, 111,
  unk, unk, unk, unk, unk, unk, unk, 29, 42, 56, 125, 97, 54, 100, 126, 164,
  166, 165, 163, 161, 115, 114, 113, 150, 158, 159, 128, 136, 177, 178, 176,
  142, 152, 173, 140, unk, unk, unk, unk
};

#undef unk

/*=========================== HID device interface ===========================*/

/** @name HID device interface */
/*@{*/
/** Handle for opened Human Interface Device */
static HIDInterface *theHid;

/** Set non-zero if ruby script called the "open device" function  */
static Bool theHidIsOpen = False;

/** Set by signal handler to quit the program. */
static Bool thePrgShouldQuit = False;

/** Number of the device endpoint.  */
static unsigned theEndPoint;


/** Type of a message */
typedef char * message_t;
/*@}*/

/*=========================== USB keyboard packets ===========================*/
/** @name USB keyboard packets
 * We hardcode the packet length here, although it is theoretically possible
 * that a USB keyboard sends other packets.
 *
 * Each USB packet consists of a byte indicating the state of the shift, ctrl,
 * alt, and windows key. The next byte is reserved.
 * Then 6 bytes follow that hold the the ids of the currently pressed keys.
 * Such a packet is sent every time a key is pressed or released.
 */
/*@{*/

/** Number of usable bytes for keycode */
#define THEPACKETLEN 6

/** Storage space for 2 packets: the last one seen and that of the current
 * event */
static char thePktBuf[2][2 + THEPACKETLEN];

/** Points to current packet */
static message_t thePacket = thePktBuf[0];

/** Points to packet of last event. */
static message_t theLastPacket = thePktBuf[1];

/** Points to the last received packet (persistent across idle events) */
static message_t theTrueLastPacket = thePktBuf[1];

/** The time of the last event. */
static struct timeval theLastTime;
/*@}*/

/*====================== ruby variables and callback ids =====================*/
/** @name ruby access
 *
 * In order to access ruby from C, we need to store function ids and pointer
 * to variables.
 */
/*@{*/

/**  Index of ruby functions ids.
  KEYDN and KEYUP need to be the 0 and 1 because they serve double duty in the
  key event processing functions
  */
typedef enum {
  /** Id of the ruby function "UHK_keyup"  */
  RFUNC_KEYDN = 0,
  /** Id of the ruby function "UHK_keydn"  */
  RFUNC_KEYUP = 1,
  /** Id of the ruby function "UHK_keyev"  */
  RFUNC_KEYEV,
  /** Id of the ruby function "UHK_mark" */
  RFUNC_MARK,
  /** Number of ruby functions */
  NUM_RUBY_FUNCS
} callbackID_t;

/** Struct to keep track function names and their respective ruby IDs */
typedef struct {
  /** Ruby ID of the function. */
  ID func;
  /** Is the function defined in any script loaded. */
  int def;
} rubyCallback_t;

/** Table of ruby function callbacks. */
static rubyCallback_t theRubyCallbacks[NUM_RUBY_FUNCS];

/** Function to setup a callback. */
static void setupRubyCallback(const char *funName, callbackID_t id)
{
  assert(id < NUM_RUBY_FUNCS);
  theRubyCallbacks[id].func = rb_intern(funName);
  theRubyCallbacks[id].def =
      rb_obj_respond_to(rb_mKernel, theRubyCallbacks[id].func, 1);
}

/** Call a ruby method from C - do nothing if not defined */
static void callRubyFunc(callbackID_t func, int n, ...)
{
  if (theRubyCallbacks[func].def != 0) {
    VALUE *argv;
    va_list ar;
    if (n > 0) {
      int i;
      va_start(ar, n);
      argv = ALLOCA_N(VALUE, n);
      for (i = 0; i < n; ++i)
	argv[i] = va_arg(ar, VALUE);
      va_end(ar);
    } else
      argv = NULL;
    (void) rb_funcall2(rb_mKernel, theRubyCallbacks[func].func, n, argv);
  }
}

/*@}*/

/*==================================== X11 ===================================*/
/** @name X11 */
/*@{*/

/** X11 server connection */
static Display *the_display = NULL;

/** Number of milliseconds to wait after a keypress was sent. */
#define INTERKEY_DELAY  0

/*@}*/

/** @name Infrastructure */
/*@{*/

/** Displays error message and help. */
static void errorHelp(const char *msg)
{
  fprintf(stderr, "usbhotkey: %s\n", msg);
  (void)
      fputs
      ("usbhotkey: Intercept USB keyboard events and send them to X.\n"
       "USAGE: usbhotkey [OPTIONS] <script> ... \n" "\n" "Options:\n"
       "  -h\n" "  --help -- Display help.\n" "\n" "Parameters:\n"
       "  <script> -- Ruby script containing the code to run on keypress.\n"
       "              You may pass multiple scripts which are executed\n"
       "              in left-to-right order.\n", stderr);

  exit(EXIT_FAILURE);
}

/** Signal handler, set a flag to quit. */
static void quitMe(int sigNr)
{
  thePrgShouldQuit = True;
}

/** Check for a ruby exception and print some help. */
static void handle_ruby_error(int status)
{
  if (status != 0) {
    long c;
    VALUE lasterr = rb_gv_get("$!");
    VALUE message = rb_obj_as_string(lasterr);
    VALUE ary = rb_funcall(RB_ERRINFO, rb_intern("backtrace"), 0);

    if (strcmp(RSTRING_PTR(message), "exit") == 0) {
      quitMe(1);
      return;
    }
    printf("usbhotkey: %s\n", RSTRING_PTR(message));
    printf("usbhotkey: Backtrace:\n");
    for (c = 0; c < RARRAY_LEN(ary); ++c)
      printf("usbhotkey: \tfrom %s\n", RSTRING_PTR(RARRAY_PTR(ary)[c]));
    errorHelp("Error executing start script\n");
  }
}

/*@}*/

/*############################ HID device handling ###########################*/
/** @name HID device interface */
/*@{*/

/** Ruby callback to open a device given its vendor and product id.  */
static VALUE
connectHid(VALUE self,
	   VALUE vendorIdVal, VALUE productIdVal, VALUE endPointVal)
{
  HIDInterfaceMatcher matcher;
  unsigned long vendorId, productId;

  hid_return ret;

  if (theHidIsOpen == True)
    rb_raise(rb_eRuntimeError, "HID already connected.");

  vendorId = NUM2UINT(vendorIdVal);
  productId = NUM2UINT(productIdVal);
  theEndPoint = (unsigned) NUM2UINT(endPointVal);

  matcher.vendor_id = (unsigned short) vendorId;
  matcher.product_id = (unsigned short) productId;
  matcher.matcher_fn = NULL;
  matcher.custom_data = NULL;
  matcher.custom_data_length = 0;

  printf("usbhotkey: Connecting to %04lx:%04lx\n", vendorId, productId);

  ret = hid_force_open(theHid, 0, &matcher, 3);
  if (ret != HID_RET_SUCCESS)
    rb_raise(rb_eRuntimeError, "Can't connect HID.");
  theHidIsOpen = True;

  return self;
}

/** Closes HID device and X11 connection. */
static void closeAll()
{
  if (!theHid)
    return;
  if (theHidIsOpen == True) {
    /* Reset the USB device before closing the HID handle This will cause the
     * device to be reclaimed by the kernel drivers as if it was unplugged and
     * replugged.  In short - give it back to the OS.  */
    (void) usb_reset(theHid->dev_handle);
    (void) hid_close(theHid);
    theHidIsOpen = False;
  }
  hid_delete_HIDInterface(&theHid);
  (void) hid_cleanup();

  if (the_display) {
    (void) XCloseDisplay(the_display);
    the_display = NULL;
  }
}

/*@}*/

/*########################### USB packet processing ##########################*/

/** Check all keys in test that are not in set and call ruby function for
 * them. */
static void process_USB_keys(message_t test, message_t set, callbackID_t evtype)
{
  unsigned i, j;

  /* Skip the header bytes */
  test += 2;
  set += 2;

  for (i = 0; i < THEPACKETLEN; ++i) {
    if (test[i] != '\0') {
      Bool found = False;
      for (j = 0; j < THEPACKETLEN; ++j) {
	if (test[i] == set[j]) {
	  found = True;
	  break;
	}
      }
      if (found == False) {
	/* Call both the up/down function and the generic function
	   Generally only one will be defined in any given script

	   This is the reason for RFUNC_KEYDN and RFUNC_KEYUP to be fixed to 0
	   and 1: These values will be passed as parameters to UHK_USB_keyev.
	 */
	VALUE key = INT2FIX(test[i]);
	callRubyFunc(evtype, 1, key);
	callRubyFunc(RFUNC_KEYEV, 2, key, INT2FIX(evtype));
      }
    }
  }
}

/** Check all shift key bits in test that are not in set and call ruby
 * function for them. */
static void
process_USB_shift_keys(unsigned char test,
		       unsigned char set, callbackID_t evtype)
{
  unsigned char mask;
  unsigned int i;
  for (i = 0; i < 8; ++i) {
    mask = 1 << i;
    if ((test & mask) != 0u) {
      if ((set & mask) == 0u) {
	/* Call both the up/down function and the generic function
	   Generally only one will be defined in any given script

	   This is the reason for RFUNC_KEYDN and RFUNC_KEYUP to be fixed to 0
	   and 1: These values will be passed as parameters to UHK_USB_keyev.
	 */
	VALUE key = INT2FIX(224 + i);
	callRubyFunc(evtype, 1, key);
	callRubyFunc(RFUNC_KEYEV, 2, key, INT2FIX(evtype));
      }
    }
  }
}

/** Process the current packet. This function calls ruby functions which might
 * raise an exception. In this case the remainder of the function is not
 * executed. */
static VALUE handlePacket(VALUE args)
{
  /* Get the time since the last event. */
  struct timeval thisTime;
  int timeElapsed;
  gettimeofday(&thisTime, NULL);
  timeElapsed =
      1000 * (thisTime.tv_sec - theLastTime.tv_sec) + 
      (thisTime.tv_usec - theLastTime.tv_usec) / 1000;

  /* Call the mark function indicating the start of the release sequence. */
  callRubyFunc(RFUNC_MARK, 2, INT2FIX(MARK_START_RELEASE),
	       INT2FIX(timeElapsed));

  /* Call the key-up function for those keys that are down, but not in
   * packet anymore. */
  process_USB_keys(theLastPacket, thePacket, RFUNC_KEYUP);

  /* Check the release of shift keys. */
  process_USB_shift_keys(theLastPacket[0], thePacket[0], RFUNC_KEYUP);

  /* Call the mark function indicating the end of the release sequence. */
  callRubyFunc(RFUNC_MARK, 2, INT2FIX(MARK_END_RELEASE_START_PUSH),
	       INT2FIX(timeElapsed));

  /* Check the press of shift keys. */
  process_USB_shift_keys(thePacket[0], theLastPacket[0], RFUNC_KEYDN);

  /* Call the key-down function for the keys that are down. */
  process_USB_keys(thePacket, theLastPacket, RFUNC_KEYDN);

  /* Call the mark function indicating the end of the push sequence. */
  callRubyFunc(RFUNC_MARK, 2, INT2FIX(MARK_END_PUSH),
	       INT2FIX(timeElapsed));

  theLastTime = thisTime;
  return 0;
}

/*############################ X11 key generation ############################*/
/** Send X11 key press event. */
static VALUE x11_keydn(VALUE self, VALUE keycode)
{
  unsigned long kc = FIX2UINT(keycode);
#ifdef DEBUG_RUBY_IF
  printf("usbhotkey: x11_keydn_norep( %d)\n", kc);
#endif
  (void) XTestFakeKeyEvent(the_display,
			   XKeysymToKeycode(the_display, kc), True,
			   INTERKEY_DELAY);
  return 0;
}

/** Send X11 key release event. */
static VALUE x11_keyup(VALUE self, VALUE keycode)
{
  unsigned long kc = FIX2UINT(keycode);
#ifdef DEBUG_RUBY_IF
  printf("usbhotkey: x11_keyup( %d)\n", kc);
#endif
  (void) XTestFakeKeyEvent(the_display,
			   XKeysymToKeycode(the_display, kc), False,
			   INTERKEY_DELAY);
  return 0;
}

/** Send X11 key press/release event (depending on value of "up") */
static VALUE x11_keyev(VALUE self, VALUE keycode, VALUE up)
{
  unsigned long kc = FIX2UINT(keycode);
  Bool down = True;
  if (FIX2UINT(up) > 0)
    down = False;
  (void) XTestFakeKeyEvent(the_display,
			   XKeysymToKeycode(the_display, kc), down,
			   INTERKEY_DELAY);
  return 0;
}

/** Send X11 button press event */
static VALUE x11_btndn(VALUE self, VALUE button)
{
  (void) XTestFakeButtonEvent(the_display, FIX2UINT(button), True,
			      INTERKEY_DELAY);
  return 0;
}

/**  Send X11 button release event */
static VALUE x11_btnup(VALUE self, VALUE button)
{
  (void) XTestFakeButtonEvent(the_display, FIX2UINT(button), False,
			      INTERKEY_DELAY);
  return 0;
}

/** Send X11 button press/release event (depending on value of "up") */
static VALUE x11_btnev(VALUE self, VALUE button, VALUE up)
{
  Bool down = True;
  if (FIX2UINT(up) > 0)
    down = False;
  (void) XTestFakeButtonEvent(the_display, FIX2UINT(button), down,
			      INTERKEY_DELAY);
  return 0;
}

/** Get the KeySym associated with the given Keycode */
static VALUE x11_get_keysym(VALUE self, VALUE code)
{
  /* Take the raw USB HID scancode, translate it to a more traditional AT type
   * scancode using the usb_keymap table, and further translate it to a usable
   * Xkb keycode by adding 8 to it (don't ask me why Xkb couldn't just use the
   * regular AT codes).  Finally, use the result to look up the associated
   * keysym.  */
  int iCode = FIX2INT(code);
  if (0 <= iCode && iCode < USB2AT_ENTRIES)
    return
	INT2FIX(XkbKeycodeToKeysym
		(the_display, usb2at_keymap[iCode] + 8u, 0, 0));
  return INT2FIX(XkbKeycodeToKeysym(the_display, 8u, 0, 0));
}

/**  Find out if a key is pressed or not */
static VALUE uhk_is_key_pressed(VALUE self, VALUE code)
{
  char c = (char) FIX2INT(code);
  message_t p = theTrueLastPacket + 2;
  unsigned i;
  for (i = 0; i < THEPACKETLEN; ++i) {
    if (p[i] == c)
      return INT2FIX(1);
  }
  return INT2FIX(0);
}


/*################################### Main ###################################*/
int main(int argc, char * *argv)
{
  int status = 0;

  hid_return ret;

  /* Process commandline */
  int opti = 1;
  while (opti < argc) {
    if (strcmp(argv[opti], "-h") == 0 || strcmp(argv[opti], "--help") == 0)
      errorHelp("Display help text.");
    else
      break;
  }

  if (opti >= argc)
    errorHelp("No script given.");

  /* Init libhid */
  if (hid_init() != HID_RET_SUCCESS) {
    printf("usbhotkey: Could not init HID interface. Exiting.\n");
    return EXIT_FAILURE;
  }
  hid_set_debug(HID_DEBUG_NONE);
  hid_set_debug_stream(stderr);
  hid_set_usb_debug(1000);

  theHid = hid_new_HIDInterface();
  (void) atexit(closeAll);
  (void) signal(SIGINT, quitMe);
  (void) signal(SIGTERM, quitMe);

  memset(theLastPacket, 0, THEPACKETLEN + 2);

  /* Init X11 connection and Test extension. */
  the_display = XOpenDisplay(NULL);
  if (!the_display)
    errorHelp("Can't open X11 connection.");

  /* Init ruby */
  ruby_init();
  ruby_init_loadpath();

  /*  Set up the C functions to be called from ruby */
  rb_define_global_function("UHK_connect", connectHid, 3);
  rb_define_global_function("UHK_keydn", x11_keydn, 1);
  rb_define_global_function("UHK_keyup", x11_keyup, 1);
  rb_define_global_function("UHK_btndn", x11_btndn, 1);
  rb_define_global_function("UHK_btnup", x11_btnup, 1);
  rb_define_global_function("UHK_keyev", x11_keyev, 2);
  rb_define_global_function("UHK_btnev", x11_btnev, 2);
  rb_define_global_function("UHK_get_keysym", x11_get_keysym, 1);
  rb_define_global_function("UHK_is_key_pressed", uhk_is_key_pressed, 1);

  create_mark_constants();
  create_usb_constants();
  create_win_constants();

  /* Load ruby scripts */
  while (opti < argc) {
    const char *script = argv[opti];
    ++opti;
    printf("usbhotkey: Loading script %s\n", script);
    rb_load_protect(rb_str_new2(script), 0, &status);
    handle_ruby_error(status);
  }

  /* Setup the ruby callbacks if they exist in the script. */
  setupRubyCallback("UHK_USB_keyup", RFUNC_KEYUP);
  setupRubyCallback("UHK_USB_keydn", RFUNC_KEYDN);
  setupRubyCallback("UHK_USB_keyev", RFUNC_KEYEV);
  setupRubyCallback("UHK_USB_mark", RFUNC_MARK);


  /* Perform USB and X11 event processing. */
  if (theHidIsOpen == True) {
    printf("usbhotkey: Starting event loop\n");

    /* Get the current time and initialize the time of the (not-seen) last
     * event to some time in the past. */
    gettimeofday(&theLastTime, NULL);
    theLastTime.tv_sec -= 10;

    /* Event loop. */
    while (thePrgShouldQuit == False) {
      memset(thePacket, 0, THEPACKETLEN + 2);

      /* Read a packet. */
      ret =
	  hid_interrupt_read(theHid, theEndPoint, thePacket,
			     THEPACKETLEN + 2, 200);

      if (ret == HID_RET_SUCCESS) {
	theTrueLastPacket = thePacket;
	(void) XTestGrabControl(the_display, True);
#ifdef DEBUG_PACKET
	unsigned i;
	printf("usbhotkey: Packet received: ");
	for (i = 0; i < THEPACKETLEN + 2; ++i)
	  printf("%02x ", thePacket[i]);
	printf("\n");
#endif

	/* Find set differences and call ruby functions. */
	(void) rb_protect(handlePacket, 0, &status);
	handle_ruby_error(status);

	/* Switch roles of packet and last packet. */
	{
	  char *h = thePacket;
	  thePacket = theLastPacket;
	  theLastPacket = h;
	}
	(void) XTestGrabControl(the_display, False);
	(void) XSync(the_display, False);
      } else
#ifdef LIBHID_HAS_TIMEOUT
      if (ret == HID_RET_TIMEOUT)
#endif
      {
	/* Call the mark function indicating an idle event. */
	/* Get the time since the last event. */
	struct timeval thisTime;
	int timeElapsed;
	gettimeofday(&thisTime, NULL);
	timeElapsed =
	    1000 * (thisTime.tv_sec - theLastTime.tv_sec) +
	    (thisTime.tv_usec - theLastTime.tv_usec) / 1000;
	(void) XTestGrabControl(the_display, True);
	callRubyFunc(RFUNC_MARK, 2, INT2FIX(MARK_IDLE),
		     INT2FIX(timeElapsed));
	(void) XTestGrabControl(the_display, False);
	(void) XSync(the_display, False);
	theLastTime = thisTime;
      }
#ifdef LIBHID_HAS_TIMEOUT
      else
	break;
#endif
    }
  } else {
    printf("usbhotkey: No device opened by script. Exiting.\n");
  }

  closeAll();

  return EXIT_SUCCESS;
}
